//////////////////////////////////////////////////////////////////////////
// BonSDK .NET binding interface library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Diagnostics;
using System.Runtime.InteropServices;

namespace Bon
{
	/// <summary>
	/// vZbTA[LeN`擾NXłB
	/// </summary>
	[ComVisible(true)]
	[Guid("EFCCF55F-5D27-40ac-8610-D3F19B9E2813")]
	[ComDefaultInterface(typeof(IArchitecture))]
	public sealed class Architecture : IArchitecture
	{
		[StructLayout(LayoutKind.Sequential)]
		private struct SYSTEM_INFO
		{
			public ProcessorArchitecture processorArchitecture;
			ushort reserved;
			public uint pageSize;
			public IntPtr minimumApplicationAddress;
			public IntPtr maximumApplicationAddress;
			public IntPtr activeProcessorMask;
			public uint numberOfProcessors;
			public uint processorType;
			public uint allocationGranularity;
			public ushort processorLevel;
			public ushort processorRevision;
		}

		[DllImport("kernel32.dll")]
		private static extern void GetNativeSystemInfo(ref SYSTEM_INFO systemInfo);
		[DllImport("kernel32.dll")]
		private static extern IntPtr GetCurrentProcess();
		[DllImport("kernel32.dll")]
		private static extern bool IsWow64Process(IntPtr handle, ref bool wow64Process);

		private readonly ProcessorArchitecture os_;
		private readonly ProcessorArchitecture currentProcess_;

		/// <summary>
		/// RXgN^łB
		/// </summary>
		public Architecture()
		{
			SYSTEM_INFO systemInfo = new SYSTEM_INFO();
			GetNativeSystemInfo(ref systemInfo);
			os_ = systemInfo.processorArchitecture;

			switch (systemInfo.processorArchitecture)
			{
				case ProcessorArchitecture.INTEL:
				case ProcessorArchitecture.IA32onWin64:
					currentProcess_ = ProcessorArchitecture.INTEL;
					break;
				case ProcessorArchitecture.AMD64:
					{
						bool wow64 = false;
						IsWow64Process(Process.GetCurrentProcess().Handle, ref wow64);
						if (wow64 == true)
						{
							currentProcess_ = ProcessorArchitecture.INTEL;
						}
						else
						{
							currentProcess_ = ProcessorArchitecture.AMD64;
						}
					}
					break;
				default:
					currentProcess_ = systemInfo.processorArchitecture;
					break;
			}
		}

		/// <summary>
		/// OS̃A[LeN`擾܂B
		/// </summary>
		public ProcessorArchitecture OS
		{
			get
			{
				return os_;
			}
		}

		/// <summary>
		/// ݂̃vZX̓A[LeN`擾܂B
		/// </summary>
		public ProcessorArchitecture CurrentProcess
		{
			get
			{
				return currentProcess_;
			}
		}
	}
}
