//////////////////////////////////////////////////////////////////////////
// BonSDK .NET binding interface library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Runtime.Serialization;
using System.Security.Permissions;

namespace Bon
{
	/// <summary>
	/// URLNXłB
	/// </summary>
	/// <remarks>̃NXUriNX̃bp[NXłB
	/// ܂AIUrlC^[tFCXĂ邽߁ACOMQƉ\łB</remarks>
	[Serializable]
	[TypeConverter(typeof(UriTypeConverter))]
	public sealed class Url : IUrl, ISerializable
	{
		private readonly Uri url_;

		public Url(string urlString)
		{
			url_ = new Uri(urlString);
		}

		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.SerializationFormatter)]
		private Url(SerializationInfo info, StreamingContext context)
		{
			url_ = new Uri(info.GetString("url_"));
		}

		public Url(string urlString, UriKind urlKind)
		{
			url_ = new Uri(urlString, urlKind);
		}

		public Url(Uri baseUrl, string relativeUrl)
		{
			url_ = new Uri(baseUrl, relativeUrl);
		}

		public Url(Uri baseUrl, Uri relativeUrl)
		{
			url_ = new Uri(baseUrl, relativeUrl);
		}

		public bool Equals(IUrl url)
		{
			if (url == null)
			{
				return false;
			}
			return url_.Equals(url.Uri);
		}

		[EditorBrowsable(EditorBrowsableState.Advanced)]
		public override bool Equals(object obj)
		{
			return Equals(obj as IUrl);
		}

		public override int GetHashCode()
		{
			return url_.GetHashCode();
		}

		public override string ToString()
		{
			return url_.ToString();
		}

		[EditorBrowsable(EditorBrowsableState.Never)]
		[SecurityPermission(SecurityAction.LinkDemand, Flags = SecurityPermissionFlag.SerializationFormatter)]
		void ISerializable.GetObjectData(SerializationInfo serializationInfo, StreamingContext streamingContext)
		{
			serializationInfo.AddValue("url_", url_.ToString());
		}

		public string AbsolutePath
		{
			get { return url_.AbsolutePath; }
		}

		public string AbsoluteUrl
		{
			get { return url_.AbsoluteUri; }
		}

		public string Authority
		{
			get { return url_.Authority; }
		}

		public string DnsSafeHost
		{
			get { return url_.DnsSafeHost; }
		}

		public string Fragment
		{
			get { return url_.Fragment; }
		}

		public string Host
		{
			get { return url_.Host; }
		}

		public bool IsAbsoluteUrl
		{
			get { return url_.IsAbsoluteUri; }
		}

		public bool IsDefaultPort
		{
			get { return url_.IsDefaultPort; }
		}

		public bool IsFile
		{
			get { return url_.IsFile; }
		}

		public bool IsLoopback
		{
			get { return url_.IsLoopback; }
		}

		public bool IsUnc
		{
			get { return url_.IsUnc; }
		}

		public string LocalPath
		{
			get { return url_.LocalPath; }
		}

		public string OriginalString
		{
			get { return url_.OriginalString; }
		}

		public string PathAndQuery
		{
			get { return url_.PathAndQuery; }
		}

		public int Port
		{
			get { return url_.Port; }
		}

		public string Query
		{
			get { return url_.Query; }
		}

		public string Scheme
		{
			get { return url_.Scheme; }
		}

		public string[] Segments
		{
			get { return url_.Segments; }
		}

		public bool UserEscaped
		{
			get { return url_.UserEscaped; }
		}

		public string UserInfo
		{
			get { return url_.UserInfo; }
		}

		public Uri Uri
		{
			get { return url_; }
		}
	}

	/// <summary>
	/// IUrlC^[tFCX̃CX^X𐶐NXłB
	/// </summary>
	/// <remarks>COM璼IUrlC^[tFCX̃CX^X𐶐oȂ߁ÃNXgpĐ܂B</remarks>
	[ComVisible(true)]
	[Guid("20C3B681-ADD9-4cab-BDD5-D8D1CC45731D")]
	[ComDefaultInterface(typeof(IUrlBuilder))]
	public sealed class UrlBuilder : UriBuilder, IUrlBuilder
	{
		public UrlBuilder()
		{
		}

		public UrlBuilder(string uri)
			: base(uri)
		{
		}

		public UrlBuilder(Uri uri)
			: base(uri)
		{
		}

		public UrlBuilder(string schemeName, string hostName)
			: base(schemeName, hostName)
		{
		}

		public UrlBuilder(string scheme, string host, int portNumber)
			: base(scheme, host, portNumber)
		{
		}

		public UrlBuilder(string scheme, string host, int port, string pathValue)
			: base(scheme, host, port, pathValue)
		{
		}

		public UrlBuilder(string scheme, string host, int port, string path, string extraValue)
			: base(scheme, host, port, path, extraValue)
		{
		}

		[ComVisible(false)]
		[EditorBrowsable(EditorBrowsableState.Never)]
		public new IUrl Uri
		{
			get
			{
				return new Url(base.Uri.ToString());
			}
		}

		public IUrl Url
		{
			get
			{
				return new Url(base.Uri.ToString());
			}
		}
	}
}
