//////////////////////////////////////////////////////////////////////////
// BonSDK .NET binding interop library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.


#include "stdafx.h"

#include "ManagedLock.h"
#include "DataQueue.h"
#include "RawTsData.h"
#include "Tuner.h"

namespace Bon
{
	DataQueue::DataQueue(Tuner^ parent, int programID, __int64 maxSize)
		: parent_(parent), maxSize_(maxSize), totalSize_(0), programID_(programID),
		queue_(gcnew System::Collections::Generic::Queue<ITsData^>()),
		queued_(gcnew System::Threading::ManualResetEvent(false))
	{
		if (programID_ < -1)
		{
			throw gcnew BonException("Invalid program ID: ID={0}", programID_);
		}
	}

	DataQueue::~DataQueue()
	{
		Internal::ManagedLock lock(this);

		if (parent_ != nullptr)
		{
			parent_->RemoveQueue(this);
			parent_ = nullptr;
		}

		Close();
	}

	void DataQueue::Close()
	{
		Internal::ManagedLock lock(queue_);

		queue_->Clear();
		queued_->Set();
		totalSize_ = 0;
	}

	void DataQueue::Clear()
	{
		Internal::ManagedLock lock(queue_);

		queue_->Clear();
		if (parent_ != nullptr)
		{
			queued_->Reset();
		}
		totalSize_ = 0;
	}

	System::Threading::WaitHandle^ DataQueue::CanDequeue::get()
	{
		return queued_;
	}

	System::IntPtr DataQueue::CanDequeueHandle::get()
	{
		return queued_->Handle;
	}

	int DataQueue::Count::get()
	{
		Internal::ManagedLock lock(queue_);

		return queue_->Count;
	}

	__int64 DataQueue::MaxSize::get()
	{
		return maxSize_;
	}

	__int64 DataQueue::TotalSize::get()
	{
		Internal::ManagedLock lock(queue_);

		return totalSize_;
	}

	int DataQueue::ProgramID::get()
	{
		Internal::ManagedLock lock(queue_);

		return programID_;
	}

	void DataQueue::ProgramID::set(int programID)
	{
		// ÕvOID̕ύX͂̃vpeBōsÃt@T[ĥ悤Ȃ̂
		// ۂɂStreamSink::ChangeProgramIDŏB
		// Iɂ͂₱AO猩DataQueueɑ΂ăvOIDݒ肵AΉ
		// Xg[oƂт삪B
		// DataQueue::ProgramID::set --> StreamSink::ChangeProgramID --> DataQueue::SetProgramID
		Internal::ManagedLock lock(queue_);

		if (programID < -1)
		{
			throw gcnew BonException("Invalid program ID: ID={0}", programID);
		}

		parent_->ChangeProgramID(this, programID);
	}

	void DataQueue::SetProgramID(int programID)
	{
		// ProgramID::setStreamSink::ChangeProgramIDoRĊԐړIɌĂяoB
		Internal::ManagedLock lock(queue_);

		programID_ = programID;
		if (queue_->Count > 0)
		{
			queue_->Clear();
			queued_->Reset();
			totalSize_ = 0;
		}
	}

	void DataQueue::Enqueue(ITsData^ data)
	{
		Internal::ManagedLock lock(queue_);

		int size = data->Size;
		if ((totalSize_ + size) > maxSize_)
		{
			BON_TRACE(TraceType_Warning, TEXT("Overflow DataQueue's queue: size=%1!d!, maxSize=%2!d!"), totalSize_ + size, maxSize_);

#ifndef NO_CLEAR_OVERFLOW
			Clear();
#else
			while (queue_->Count > 0)
			{
				totalSize_ -= queue_->Dequeue()->Size;
				if ((totalSize_ + size) <= maxSize_)
				{
					break;
				}
			}
#endif
		}

		queue_->Enqueue(data);
		if (queue_->Count == 1)
		{
			queued_->Set();
		}
		totalSize_ += size;
	}

	ITsData^ DataQueue::Dequeue()
	{
		queued_->WaitOne();

		Internal::ManagedLock lock(queue_);

		if (queue_->Count == 0)
		{
			throw gcnew System::ObjectDisposedException("DataQueue disposed");
		}

		ITsData^ data = queue_->Dequeue();
		if (queue_->Count == 0)
		{
			queued_->Reset();
		}
		totalSize_ -= data->Size;

		return data;
	}

	IRawTsData^ DataQueue::RawDequeue()
	{
		ITsData^ data = Dequeue();
		return gcnew RawTsData(data);
	}
}
