//////////////////////////////////////////////////////////////////////////
// BonSDK .NET binding interop library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.


#include "stdafx.h"

#include "CreateBonInstance.h"
#include "ManagedLock.h"
#include "StreamSink.h"
#include "TsData.h"
#include "DataQueueManager.h"
#include "DataQueue.h"

namespace Bon
{
	namespace Internal
	{
		StreamSink::StreamSink(IBonObject *pOwner, DWORD bufferSize)
			: CMediaDecoder(pOwner), bufferSize_(bufferSize), queueCount_(0), programIDVersion_(0),
			managerList_(gcnew System::Collections::Generic::Dictionary<int, DataQueueManager^>()),
			pProgAnalyzer_(dynamic_cast<const IProgAnalyzer *>(pOwner))
		{
		}

		StreamSink::~StreamSink()
		{
			ManagedLock lock(managerList_);

			for each (System::Collections::Generic::KeyValuePair<int, DataQueueManager^> manager in
				static_cast<System::Collections::Generic::Dictionary<int, DataQueueManager^>^>(managerList_))
			{
				manager.Value->~DataQueueManager();
			}

			managerList_->Clear();
		}

		const bool StreamSink::OnPlay()
		{
			Clear();
			return CMediaDecoder::OnPlay();
		}

		const bool StreamSink::OnStop()
		{
			Clear();
			return CMediaDecoder::OnStop();
		}

		const bool StreamSink::OnReset()
		{
			Clear();
			return CMediaDecoder::OnReset();
		}

		const bool StreamSink::OnInputMedia(const IMediaData *pMediaData, const DWORD dwInputIndex)
		{
			const BYTE *pData = pMediaData->GetData();
			DWORD dataSize = pMediaData->GetSize();

			Internal::ManagedLock lock(managerList_);

			// ̃NX͂ŃpPbg̑IʂsADataQueueProgramIDݒɏ]
			// `肷B
			// ProgramID=-1̏ꍇRaw`ƂASẴpPbg`B
			// ProgramID=0̏ꍇ͖ݒƂAIProgAnalyzerJgProgramIDw肳Ă̂ƂB
			// TODO:nbVProgramID`肷ƂACfA͊SɎsB
			//   ̂߁Aςɂ₱ȂĂ܂B

			// DataQueue蓖ĂĂȂΖ
			if (queueCount_ == 0)
			{
				return true;
			}

			// ProgramID-1ɐݒ肳ĂDataQueue΁ASẴpPbg`
			// iRaw`[hj
			DataQueueManager^ manager = nullptr;
			if (managerList_->TryGetValue(-1, manager) == true)
			{
				manager->Enqueue(pData, dataSize);
			}

			// iÔ߁j
			if (pProgAnalyzer_ == 0)
			{
				return true;
			}

			// iÔ߁j
			const ITsPacket *pPacket = dynamic_cast<const ITsPacket *>(pMediaData);
			if (pPacket == 0)
			{
				return true;
			}

			// pPbgID擾
			WORD pid = pPacket->GetPID();
			switch (pid)
			{
			// PAT
			case 0x00:
				// ProgramIDɑΉPAT𐶐
				{
					// PATō
					BYTE newPatData[TS_PACKET_SIZE];
					memset(newPatData, 0xff, sizeof newPatData);

					newPatData[0] = 0x47;	// [h
					newPatData[1] = 0x40;	// Payload start
					newPatData[2] = 0x00;	// PID=0
					newPatData[3] = 0x10 | pPacket->m_Header.byContinuityCounter;	// No adaptation field

					newPatData[4] = 0x00;	// (Staffing byte)

					// Section
					newPatData[5] = 0x00;	// Table ID
					newPatData[6] = 0xb0;	// Section syntax indicator / Fixed bits
					newPatData[7] = 5 + 12;	// Size

					// Extended section
					WORD tsID = pProgAnalyzer_->GetTsID();
					newPatData[8] = static_cast<BYTE>(tsID >> 8);	// Table ID Extension (TS ID)
					newPatData[9] = static_cast<BYTE>(tsID & 0xff);
					newPatData[10] = 0xc1 | ((programIDVersion_ & 0x1f) << 1);	// Fixed bits / Version / Current next indicator
					newPatData[11] = 0x00;	// Section number
					newPatData[12] = 0x00;	// Last section number

					// Payload
					newPatData[13] = 0x00;	// NIT Program ID
					newPatData[14] = 0x00;
					newPatData[15] = 0x00;	// NIT PID
					newPatData[16] = 0x10;

					// PATSĂDataQueueManagerɓ`
					for each (System::Collections::Generic::KeyValuePair<int, DataQueueManager^> entry in
						static_cast<System::Collections::Generic::Dictionary<int, DataQueueManager^>^>(managerList_))
					{
						// ProgramID֘AtĂȂDataQueueManagerɂ͓`ς݂Ȃ̂ŏO
						WORD programID = static_cast<WORD>(entry.Key);
						if (programID != -1)
						{
							// ProgramIDݒ̏ꍇ́AJggp
							if (programID == 0)
							{
								programID = pProgAnalyzer_->GetCurrentProgramID();
							}

							// ProgramIDPmt PID𓾂iFĂȂ0j
							WORD pmtPID = pProgAnalyzer_->GetPmtPIDFromProgramID(programID);
							if (pmtPID > 0)
							{
								// PATɏݒ
								newPatData[17] = static_cast<BYTE>(programID >> 8);	// Program ID
								newPatData[18] = static_cast<BYTE>(programID & 0xff);
								newPatData[19] = static_cast<BYTE>(pmtPID >> 8);	// PmtPID
								newPatData[20] = static_cast<BYTE>(pmtPID & 0xff);

								// CRCvZ
								DWORD crc = CALCULATE_CRC(newPatData + 5, 16, 0xffffffffUL);
								newPatData[21] = static_cast<BYTE>(crc >> 24);
								newPatData[22] = static_cast<BYTE>(crc >> 16);
								newPatData[23] = static_cast<BYTE>(crc >> 8);
								newPatData[24] = static_cast<BYTE>(crc & 0xff);

								// PATL[ɓ
								entry.Value->Enqueue(newPatData, sizeof newPatData);
//								entry.Value->Enqueue(pData, dataSize);
							}
						}
					}
				}
				break;
			// CAT, NIT, SDT, EIT
			case 0x01:
			case 0x10:
			case 0x11:
			case 0x12:
				// vO肵Ă̂Ŗ
				// iProgramID֘AtĂȂDataQueueManagerɂ͓`ς݁j
				break;
			// TOT
			case 0x14:
				// TOT͑SĂDataQueueManagerɓ`
				for each (System::Collections::Generic::KeyValuePair<int, DataQueueManager^> entry in
					static_cast<System::Collections::Generic::Dictionary<int, DataQueueManager^>^>(managerList_))
				{
					// ProgramID֘AtĂȂDataQueueManagerɂ͓`ς݂Ȃ̂ŏO
					int programID = entry.Key;
					if (programID != -1)
					{
						// L[ɓ
						entry.Value->Enqueue(pData, dataSize);
					}
				}
				break;
			// ̑PID
			default:
				{
					// ProgramID݂pPbĝݓ`
					WORD programID = pProgAnalyzer_->GetProgramIDFromPID(pid);
					if (programID != 0)
					{
						// ProgramID֘AtĂDataQueueManagerɓ`
						// iProgramID֘AtĂȂDataQueueManagerɂ͓`ς݁j
						if (managerList_->TryGetValue(programID, manager) == true)
						{
							// L[ɓ
							manager->Enqueue(pData, dataSize);
						}

						// ProgramIDݒ(0)̃L[ɓ`
						if (programID == pProgAnalyzer_->GetCurrentProgramID())
						{
							if (managerList_->TryGetValue(0, manager) == true)
							{
								// L[ɓ
								manager->Enqueue(pData, dataSize);
							}
						}
					}
				}
				break;
			}

			return true;
		}

		void StreamSink::Clear()
		{
			Internal::ManagedLock lock(managerList_);

			// SĂDataQueueManagerNAǐʑSĂDataQueueNAj
			for each (System::Collections::Generic::KeyValuePair<int, DataQueueManager^> manager in
				static_cast<System::Collections::Generic::Dictionary<int, DataQueueManager^>^>(managerList_))
			{
				manager.Value->Clear();
			}
		}

		void StreamSink::InternalAddQueue(DataQueue^ queue, int programID)
		{
			// ProgramIDɑΉDataQueueManager܂蓖ĂĂȂ
			DataQueueManager^ manager = nullptr;
			if (managerList_->TryGetValue(programID, manager) == false)
			{
				// 蓖Ă
				manager = gcnew DataQueueManager(bufferSize_);
				managerList_->Add(programID, manager);
			}

			// DataQueueManagerDataQueueo^
			manager->RegisterQueue(queue);
		}

		void StreamSink::InternalRemoveQueue(DataQueue^ queue, int programID)
		{
			// DataQueueManagerDataQueueO
			DataQueueManager^ manager = managerList_->default::get(programID);
			if (manager->UnregisterQueue(queue) == 0)
			{
				// DataQueueManagerDataQueueЂƂȂȂADataQueueManager̂폜
				manager->~DataQueueManager();
				managerList_->Remove(programID);
			}
		}

		System::Collections::Generic::KeyValuePair<IDataQueue^, int>
			StreamSink::CreateQueue(Tuner^ parent, int programID, __int64 maxSize)
		{
			DataQueue^ queue = gcnew DataQueue(parent, programID, maxSize);

			Internal::ManagedLock lock(managerList_);

			InternalAddQueue(queue, programID);
			queueCount_++;

			return System::Collections::Generic::KeyValuePair<IDataQueue^, int>(queue, queueCount_);
		}

		void StreamSink::ChangeProgramID(DataQueue^ queue, int newProgramID)
		{
			Internal::ManagedLock lock(managerList_);

			InternalRemoveQueue(queue, queue->ProgramID);
			InternalAddQueue(queue, newProgramID);
			queue->SetProgramID(newProgramID);

			programIDVersion_++;
		}

		void StreamSink::ChangeProgramIDForAllQueue(int newProgramID)
		{
			Internal::ManagedLock lock(managerList_);

			for each (DataQueueManager^ manager in
				gcnew System::Collections::Generic::List<DataQueueManager^>(managerList_->Values))
			{
				for each (DataQueue^ queue in
					gcnew System::Collections::Generic::List<DataQueue^>(manager))
				{
					InternalRemoveQueue(queue, queue->ProgramID);
					InternalAddQueue(queue, newProgramID);
					queue->SetProgramID(newProgramID);
				}
			}

			programIDVersion_++;
		}

		int StreamSink::RemoveQueue(DataQueue^ queue)
		{
			Internal::ManagedLock lock(managerList_);

			InternalRemoveQueue(queue, queue->ProgramID);
			queueCount_--;

			return queueCount_;
		}
	}
}
