﻿//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Diagnostics;
using System.Globalization;
using System.Runtime.InteropServices;
using System.ServiceProcess;
using System.Text;
using System.Threading;
using System.Windows.Forms;

namespace Bon
{
	public static class Program
	{
		private static bool isService_ = false;

		/// <summary>
		/// アプリケーションのメイン エントリ ポイントです。
		/// </summary>
		[STAThread]
		public static void Main(string[] args)
		{
			Thread.CurrentThread.Name = "StreamService.Main";

			try
			{
				TunerFactory.SetTraceCallback(BonTraceCallback);
				ProcessArguments(args);
			}
			catch (Exception ex)
			{
				Logger.Write(ex);

				if (isService_ == false)
				{
					using (ExceptionForm form = new ExceptionForm(ex))
					{
						form.ShowDialog();
					}
				}
			}
		}

		private static void BonTraceCallback(TraceType type, string message)
		{
			Logger.Write(type, message);
		}

		/// <summary>
		/// 指定された引数群を処理します。
		/// </summary>
		/// <param name="args">引数群</param>
		private static void ProcessArguments(string[] args)
		{
			if (args.Length == 0)
			{
				ShowSettingDialog();
				return;
			}

			string arg0 = args[0].ToLower();
			switch (arg0)
			{
				case "/service":
					isService_ = true;
					using (StreamService service = new StreamService())
					{
						ServiceBase.Run(new ServiceBase[] { service });
					}
					break;
				case "/start":
					using (ServiceController controller = new ServiceController(StreamService.ServiceName))
					{
						if (controller.Status == ServiceControllerStatus.Stopped)
						{
							controller.Start();
							controller.WaitForStatus(ServiceControllerStatus.Running, new TimeSpan(0, 0, 30));
							if (controller.Status != ServiceControllerStatus.Running)
							{
								MessageBox.Show("Cannot start streaming service.",
									"BonSDK MPEG-TS Streaming service",
									MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
							}
						}
						else
						{
							MessageBox.Show("Streaming service is already running.",
								"BonSDK MPEG-TS Streaming service",
								MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
						}
					}
					break;
				case "/stop":
					using (ServiceController service = new ServiceController(StreamService.ServiceName))
					{
						if (service.Status != ServiceControllerStatus.Stopped)
						{
							service.Stop();
							service.WaitForStatus(ServiceControllerStatus.Stopped, new TimeSpan(0, 0, 30));
							if (service.Status != ServiceControllerStatus.Stopped)
							{
								MessageBox.Show("Cannot stop streaming service.",
									"BonSDK MPEG-TS Streaming service",
									MessageBoxButtons.OK, MessageBoxIcon.Exclamation);
							}
						}
					}
					break;
				case "/debug":
					using (StreamManager manager = new StreamManager())
					{
						manager.Start();

						MessageBox.Show("Streaming service is manually running ...",
							"BonSDK MPEG-TS Streaming service",
							MessageBoxButtons.OK, MessageBoxIcon.Information);

						manager.Stop();
					}
					break;
				default:
					MessageBox.Show(
						"Invalid argument.\r\nusage: Bon.StreamService.exe [/{start|stop|debug}]",
						"BonSDK MPEG-TS Streaming service",
						MessageBoxButtons.OK, MessageBoxIcon.Error);
					break;
			}
		}

		private static void ShowSettingDialog()
		{
			Application.EnableVisualStyles();
			Application.SetCompatibleTextRenderingDefault(false);
			Application.ThreadException += new ThreadExceptionEventHandler(Application_ThreadException);
			Application.Run(new SettingForm());
		}

		private static void Application_ThreadException(object sender, ThreadExceptionEventArgs e)
		{
			using (ExceptionForm form = new ExceptionForm(e.Exception))
			{
				Form rootForm = null;
				if (Application.OpenForms.Count > 0)
				{
					rootForm = Application.OpenForms[0];
					while (true)
					{
						Form parent = rootForm.Parent as Form;
						if (parent == null)
						{
							break;
						}
						rootForm = parent;
					}
					if (rootForm.IsDisposed == true)
					{
						rootForm = null;
					}
				}
				form.ShowDialog(rootForm);
			}
		}
	}
}
