//////////////////////////////////////////////////////////////////////////
// BonSDK Stream service
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

using System;
using System.Collections;
using System.Collections.Generic;
using System.IO;
using System.Net;
using System.Runtime.Remoting.Lifetime;
using System.Runtime.Serialization;
using System.Runtime.Serialization.Formatters;
using System.Runtime.Serialization.Formatters.Soap;
using System.Security.Permissions;
using System.Threading;
using Bon.Remoting;
using System.Text;

namespace Bon
{
	/// <summary>
	/// R[_[̎NXłB
	/// </summary>
	internal sealed class RecorderManager : MarshalByRefObject, IRemoteRecorder
	{
		private static readonly TimeSpan infinite_ = new TimeSpan(0, 0, 0, 0, Timeout.Infinite);

		private readonly string dataBasePath_;
		private readonly IDictionary<string, string> storagePaths_;
		private readonly long mininumSpaceRequirements_;
		private readonly int bufferSize_;
		private readonly TimeSpan extraTimeSpan_;
		private readonly ManualResetEvent abort_ = new ManualResetEvent(true);
		private readonly SoapFormatter formatter_ = new SoapFormatter();
		private readonly Dictionary<Guid, RegisteredProgram> programs_ = new Dictionary<Guid, RegisteredProgram>();
		private readonly Dictionary<Guid, RecordingStates> programStates_ = new Dictionary<Guid, RecordingStates>();
		private readonly SortedList<DateTime, List<RegisteredProgram>> programQueue_ = new SortedList<DateTime, List<RegisteredProgram>>();
		private readonly Dictionary<Guid, ManualResetEvent> stops_ = new Dictionary<Guid, ManualResetEvent>();
		private Timer timer_;

		/// <summary>
		/// RXgN^łB
		/// </summary>
		/// <param name="dataBasePath">R[_[̏i[f[^x[Xt@Cւ̃pX</param>
		/// <param name="storagePaths">Xg[ۑXg[Wt@C̃pXQiXg[WID/tH[}bgj</param>
		/// <param name="mininumSpaceRequirements">Xg[Wt@CۑfBXN̍ŒeʁiKoCgPʁj</param>
		/// <param name="bufferSize">Xg[ۑۂɎgpobt@̃TCY</param>
		/// <param name="extraTimeSpan">\^̎OE̎ԕ</param>
		public RecorderManager(string dataBasePath, IDictionary<string, string> storagePaths,
			long mininumSpaceRequirements, int bufferSize, TimeSpan extraTimeSpan)
		{
			dataBasePath_ = dataBasePath;
			storagePaths_ = storagePaths;
			mininumSpaceRequirements_ = mininumSpaceRequirements;
			bufferSize_ = bufferSize;
			extraTimeSpan_ = extraTimeSpan;

			formatter_.AssemblyFormat = FormatterAssemblyStyle.Simple;
			formatter_.FilterLevel = TypeFilterLevel.Full;
			formatter_.TypeFormat = FormatterTypeStyle.TypesWhenNeeded;

			timer_ = new Timer(new TimerCallback(this.TimerEntry));

			Load();
		}

		/// <summary>
		/// Ώۂ̃CX^X̗Lԃ|V[𐧌䂷ALԃT[rX IuWFNg擾܂B
		/// </summary>
		/// <returns>Ώۂ̃CX^X̗Lԃ|V[𐧌䂷ƂɎgpASystem.Runtime.Remoting.Lifetime.ILease ^̃IuWFNgB</returns>
		[SecurityPermission(SecurityAction.Demand, Flags = SecurityPermissionFlag.Infrastructure)]
		public override object InitializeLifetimeService()
		{
			ILease lease = (ILease)base.InitializeLifetimeService();
			if (lease.CurrentState == LeaseState.Initial)
			{
				lease.InitialLeaseTime = TimeSpan.Zero;
			}
			return lease;
		}

		private void AddQueue(RegisteredProgram program)
		{
			List<RegisteredProgram> queue = null;
			if (programQueue_.TryGetValue(program.StartTime, out queue) == false)
			{
				queue = new List<RegisteredProgram>();
				programQueue_.Add(program.StartTime, queue);
			}
			queue.Add(program);
		}

		private void RemoveQueue(RegisteredProgram program)
		{
			List<RegisteredProgram> queue = null;
			if (programQueue_.TryGetValue(program.StartTime, out queue) == true)
			{
				queue.Remove(program);
				if (queue.Count == 0)
				{
					programQueue_.Remove(program.StartTime);
				}
			}
		}

		/// <summary>
		/// f[^x[Xt@C[h܂B
		/// </summary>
		/// <remarks>̊֐͓Ŏgp܂B</remarks>
		private void Load()
		{
			try
			{
				using (FileStream stream = new FileStream(dataBasePath_, FileMode.Open, FileAccess.Read, FileShare.Read))
				{
					RegisteredProgramPair[] programPairs = formatter_.Deserialize(stream) as RegisteredProgramPair[];
					if (programPairs == null)
					{
						return;
					}

					DateTime now = DateTime.Now;

					// tVAʂo^
					programs_.Clear();
					programStates_.Clear();
					programQueue_.Clear();
					for (int i = 0; i < programPairs.Length; i++)
					{
						// \񃊃Xgɉ
						RegisteredProgramPair pair = programPairs[i];
						RegisteredProgram program = pair.Program;
						RecordingStates status = pair.RecordingStatus;
						programs_.Add(program.RegisteredID, program);
						programStates_.Add(program.RegisteredID, status);

						// ^Jn҂ԂȂ
						if (status == RecordingStates.Waiting)
						{
							// ^C}[L[ɓo^
							AddQueue(program);
						}
					}
				}
			}
			catch
			{
			}
		}

		/// <summary>
		/// f[^x[Xt@CɌ݂̗\QZ[u܂B
		/// </summary>
		private void Save()
		{
			string folderPath = Path.GetDirectoryName(dataBasePath_);
			if (Directory.Exists(folderPath) == false)
			{
				Directory.CreateDirectory(folderPath);
			}

			string tempPath = dataBasePath_ + ".tmp";
			try
			{
				using (FileStream stream = new FileStream(tempPath, FileMode.Create, FileAccess.ReadWrite, FileShare.None))
				{
					List<RegisteredProgramPair> programPairs = new List<RegisteredProgramPair>();
					foreach (RegisteredProgram program in programs_.Values)
					{
						programPairs.Add(new RegisteredProgramPair(program, programStates_[program.RegisteredID]));
					}
					formatter_.Serialize(stream, programPairs.ToArray());
					stream.Close();
				}

				File.Delete(dataBasePath_);
				File.Move(tempPath, dataBasePath_);
			}
			catch
			{
				File.Delete(tempPath);
				throw;
			}
		}

		/// <summary>
		/// \^擮Jn܂B
		/// </summary>
		internal void Start()
		{
			lock (this)
			{
				abort_.Reset();
				StartTimer();
			}
		}

		/// <summary>
		/// \^擮~܂B
		/// </summary>
		internal void Stop()
		{
			lock (this)
			{
				timer_.Change(infinite_, infinite_);
				abort_.Set();
			}

			// ܂ꂩĂ܂B܂...
			while (true)
			{
				lock (this)
				{
					if (stops_.Count == 0)
					{
						break;
					}
				}
				Thread.Sleep(500);
			}

			lock (this)
			{
				Save();
			}
		}

		/// <summary>
		/// ^C}[JnAL[ɕۑĂ\so悤ɂ܂B
		/// </summary>
		/// <remarks>̊֐͓Ŏgp܂B</remarks>
		private void StartTimer()
		{
			if (programQueue_.Count > 0)
			{
				DateTime now = DateTime.Now;
				DateTime next = programQueue_.Keys[0] - extraTimeSpan_;
				if (next < now)
				{
					next = now;
				}
				timer_.Change(next - now, infinite_);
			}
		}

		/// <summary>
		/// w肳ꂽ\ɏ]āA^Jn܂B
		/// </summary>
		/// <param name="program">\</param>
		/// <param name="isManual">蓮ŊJnꍇtrue</param>
		/// <remarks>̊֐͓Ŏgp܂B</remarks>
		private void StartRecord(RegisteredProgram program, bool isManual)
		{
			RecordingStates state = programStates_[program.RegisteredID];
			if ((state == RecordingStates.Waiting) || (state == RecordingStates.RecordingStoppedForManual) ||
				(state == RecordingStates.RecordingAborted) || (state == RecordingStates.RecordingFnished))
			{
				if (isManual == true)
				{
					programStates_[program.RegisteredID] = RecordingStates.RecordingForManualStart;
				}
				else
				{
					programStates_[program.RegisteredID] = RecordingStates.RecordingForRegisterdProgram;
				}

				RemoveQueue(program);

				Thread thread = new Thread(new ParameterizedThreadStart(this.RecordingEntry));
				thread.Start(new KeyValuePair<RegisteredProgram, bool>(program, isManual));
			}
		}

		/// <summary>
		/// w肳ꂽ\ɑΉ^~܂B
		/// </summary>
		/// <param name="program">\</param>
		/// <remarks>̊֐͓Ŏgp܂B</remarks>
		private void StopRecord(RegisteredProgram program)
		{
			ManualResetEvent stop = null;
			if (stops_.TryGetValue(program.RegisteredID, out stop) == true)
			{
				stop.Set();
			}
		}

		/// <summary>
		/// w肳ꂽ\폜܂B
		/// </summary>
		/// <remarks>̊֐͓Ŏgp܂B</remarks>
		/// <param name="program">\</param>
		private void Remove(RegisteredProgram program)
		{
			programs_.Remove(program.RegisteredID);
			programStates_.Remove(program.RegisteredID);

			RemoveQueue(program);

			StartTimer();
		}

		/// <summary>
		/// R[_[̗LԂ擾܂B
		/// </summary>
		public bool Enabled
		{
			get
			{
				return abort_.WaitOne(0) == false;
			}
		}

		/// <summary>
		/// \Q̐擾܂B
		/// </summary>
		public int Count
		{
			get
			{
				lock (this)
				{
					return programs_.Count;
				}
			}
		}

		/// <summary>
		/// w肳ꂽ\ID̗\擾܂B
		/// </summary>
		/// <param name="registeredID">\ID</param>
		/// <returns>\</returns>
		public IRegisteredProgram this[Guid registeredID]
		{
			get
			{
				lock (this)
				{
					RegisteredProgram program = null;
					if (programs_.TryGetValue(registeredID, out program) == false)
					{
						throw new ArgumentException("Program not registered.");
					}
					return program;
				}
			}
		}

		/// <summary>
		/// t@C̕ۑΏۂƂȂXg[W̍Œ󂫗eʂ擾܂B
		/// </summary>
		public long MinimumSpaceRequirements
		{
			get
			{
				return mininumSpaceRequirements_;
			}
		}

		/// <summary>
		/// t@C̕ۑΏۂƂȂXg[WIDQ擾܂B
		/// </summary>
		public string[] StorageIDs
		{
			get
			{
				return new List<string>(storagePaths_.Keys).ToArray();
			}
		}

		/// <summary>
		/// Xg[W݂̌̋󂫗eʂ擾܂B
		/// </summary>
		/// <param name="storageID">Xg[i[Xg[WID</param>
		/// <returns>󂫗e</returns>
		public long GetCurrentSpaces(string storageID)
		{
			string storagePath = null;
			if (storagePaths_.TryGetValue(storageID, out storagePath) == false)
			{
				throw new ArgumentException("Invalid storage ID.");
			}
			string root = Directory.GetDirectoryRoot(storagePath);
			try
			{
				DriveInfo di = new DriveInfo(root);
				return di.AvailableFreeSpace;
			}
			catch
			{
				return 0;
			}
		}

		private void WriteLog(RegisteredProgram program, string message)
		{
			Logger.Write("{0}: {1}\r\n Stream URL={2}\r\n Start time={3} [{4}]\r\n Duration={5}\r\n Tuning space={6}:{7}:{8}",
				message, program.RegisteredID, program.TunerStreamUrl,
				program.StartTime, program.StartTime.ToUniversalTime(), program.Duration,
				program.TuningSpace, program.Channel, program.ProgramID);
		}

		private void WriteLog(RegisteredProgram program, string message,
			string filePath, DateTime realStartTime, TimeSpan realDuration, long currentSpaces)
		{
			Logger.Write("{0}: {1}\r\n Stream URL={2}\r\n File path={3}\r\n Start time={4} [{5}]\r\n Duration={6}\r\n Real start time={7} [{8}]\r\n Real duration={9}\r\n Tuning space={10}:{11}:{12}\r\n Storage spaces={13} / {14}",
				message, program.RegisteredID, program.TunerStreamUrl, filePath,
				program.StartTime, program.StartTime.ToUniversalTime(), program.Duration,
				realStartTime, realStartTime.ToUniversalTime(), realDuration,
				program.TuningSpace, program.Channel, program.ProgramID, currentSpaces, mininumSpaceRequirements_);
		}

		/// <summary>
		/// vO\s܂B
		/// </summary>
		/// <param name="tunerStreamUrl">`[i[Xg[URL</param>
		/// <param name="tuningSpace">`[i[Xy[X</param>
		/// <param name="channel">`l</param>
		/// <param name="programID">vOID</param>
		/// <param name="startTime">Jn(UTC)</param>
		/// <param name="duration">Ԋu</param>
		/// <param name="storageID">Xg[i[Xg[WID</param>
		/// <param name="fileName">t@C</param>
		/// <returns>\ID</returns>
		/// <remarks>t@Cɂ͗\ɊÂvtBbNXǉ܂B</remarks>
		public Guid RegisterProgram(IUrl tunerStreamUrl, int tuningSpace, int channel, int programID,
			DateTime startTime, TimeSpan duration, string storageID, string fileName)
		{
			if (storagePaths_.ContainsKey(storageID) == false)
			{
				throw new ArgumentException("Invalid storage ID.");
			}
			if (string.IsNullOrEmpty(fileName) == true)
			{
				throw new ArgumentException("Invalid file name.");
			}

			RegisteredProgram program = new RegisteredProgram(
				tunerStreamUrl, tuningSpace, channel, programID, startTime, duration, storageID, fileName);

			lock (this)
			{
				programs_.Add(program.RegisteredID, program);
				programStates_.Add(program.RegisteredID, RecordingStates.Waiting);

				AddQueue(program);

				WriteLog(program, "Program registered");

				StartTimer();
			}

			return program.RegisteredID;
		}

		/// <summary>
		/// ^Jn܂B
		/// </summary>
		/// <param name="tunerStreamUrl">`[i[Xg[URL</param>
		/// <param name="tuningSpace">`[i[Xy[X</param>
		/// <param name="channel">`l</param>
		/// <param name="programID">vOID</param>
		/// <param name="storageID">Xg[i[Xg[WID</param>
		/// <param name="fileName">t@C</param>
		/// <returns>\ID</returns>
		/// <remarks>t@Cɂ͗\ɊÂvtBbNXǉ܂B</remarks>
		public Guid StartRecord(IUrl tunerStreamUrl, int tuningSpace, int channel, int programID,
			string storageID, string fileName)
		{
			if (storagePaths_.ContainsKey(storageID) == false)
			{
				throw new ArgumentException("Invalid storage ID.");
			}
			if (string.IsNullOrEmpty(fileName) == true)
			{
				throw new ArgumentException("Invalid file name.");
			}

			RegisteredProgram program = new RegisteredProgram(
				tunerStreamUrl, tuningSpace, channel, programID,
				DateTime.Now, new TimeSpan(1, 0, 0), storageID, fileName);

			lock (this)
			{
				programs_.Add(program.RegisteredID, program);
				programStates_.Add(program.RegisteredID, RecordingStates.Waiting);

				// ɊJnꍇłAUL[ɕۑ
				AddQueue(program);

				WriteLog(program, "Program registered and start recording");

				StartRecord(program, true);
			}

			return program.RegisteredID;
		}

		/// <summary>
		/// \^̏Ԃ擾܂B
		/// </summary>
		/// <param name="registeredID">\ID</param>
		/// <returns>\</returns>
		public RecordingStates GetRecordingStatus(Guid registeredID)
		{
			lock (this)
			{
				RecordingStates status = RecordingStates.Waiting;
				if (programStates_.TryGetValue(registeredID, out status) == false)
				{
					throw new ArgumentException("Program not registered.");
				}
				return status;
			}
		}

		/// <summary>
		/// ^IɊJn܂B
		/// </summary>
		/// <param name="registeredID">\ID</param>
		public void StartRecord(Guid registeredID)
		{
			lock (this)
			{
				RegisteredProgram program = null;
				if (programs_.TryGetValue(registeredID, out program) == false)
				{
					throw new ArgumentException("Program not registered.");
				}
				StartRecord(program, true);
				StartTimer();
			}
		}

		/// <summary>
		/// ^JnƊԊuĐݒ肵܂B
		/// </summary>
		/// <param name="registeredID">\ID</param>
		/// <param name="startTime">^Jn</param>
		/// <param name="duration">^Ԋu</param>
		/// <remarks>̃\bhgpƁA^JnƊԊuύX鎖o܂B
		/// ̍ہA蓮ŊJn^̏ꍇ͗\^ƓɕύX܂B</remarks>
		public void ResetTime(Guid registeredID, DateTime startTime, TimeSpan duration)
		{
			lock (this)
			{
				RegisteredProgram program = null;
				if (programs_.TryGetValue(registeredID, out program) == false)
				{
					throw new ArgumentException("Program not registered.");
				}

				lock (program)
				{
					RecordingStates state = programStates_[program.RegisteredID];
					switch (state)
					{
						case RecordingStates.Waiting:
							RemoveQueue(program);
							program.StartTime = startTime;
							program.Duration = duration;
							AddQueue(program);
							break;
						case RecordingStates.RecordingForManualStart:
						case RecordingStates.RecordingForRegisterdProgram:
							program.StartTime = startTime;
							program.Duration = duration;
							break;
						case RecordingStates.RecordingStoppedForManual:
						case RecordingStates.RecordingAborted:
						case RecordingStates.RecordingFnished:
							program.StartTime = startTime;
							program.Duration = duration;
							programStates_[program.RegisteredID] = RecordingStates.Waiting;
							AddQueue(program);
							DateTime now = DateTime.Now;
							if (now >= startTime)
							{
								StartRecord(program, true);
							}
							else
							{
								StartTimer();
							}
							break;
					}

					WriteLog(program, "Reset program time");
				}
			}
		}

		/// <summary>
		/// ^Iɒ~܂B
		/// </summary>
		/// <param name="registeredID">\ID</param>
		public void StopRecord(Guid registeredID)
		{
			RegisteredProgram program = null;
			lock (this)
			{
				if (programs_.TryGetValue(registeredID, out program) == false)
				{
					throw new ArgumentException("Program not registered.");
				}
				StopRecord(program);
			}
			while (true)
			{
				lock (this)
				{
					if (stops_.ContainsKey(program.RegisteredID) == false)
					{
						break;
					}
				}
				Thread.Sleep(500);
			}
		}

		/// <summary>
		/// ̗\폜܂B
		/// </summary>
		/// <param name="registeredID">\ID</param>
		public void Remove(Guid registeredID)
		{
			RegisteredProgram program = null;
			lock (this)
			{
				if (programs_.TryGetValue(registeredID, out program) == false)
				{
					throw new ArgumentException("Program not registered.");
				}
				StopRecord(program);
			}
			while (true)
			{
				lock (this)
				{
					if (stops_.ContainsKey(program.RegisteredID) == false)
					{
						break;
					}
				}
				Thread.Sleep(500);
			}
			lock (this)
			{
				Remove(program);
			}

			WriteLog(program, "Program removed");
		}

		/// <summary>
		/// \Q񋓂܂B
		/// </summary>
		/// <returns>񋓎q</returns>
		public IEnumerator GetEnumerator()
		{
			lock (this)
			{
				return new List<RegisteredProgram>(programs_.Values).ToArray().GetEnumerator();
			}
		}

		/// <summary>
		/// ^Jn^C}[̃Gg|CgłB
		/// </summary>
		/// <param name="state"></param>
		/// <remarks>vOL[QƂAԂo߂ĂΘ^揈JniXbhNj܂B</remarks>
		private void TimerEntry(object state)
		{
			lock (this)
			{
				if (programQueue_.Count > 0)
				{
					// ^C}[oߎԂƌݎvĂȂ\̂ŁAÔ߂ɊmF
					DateTime now = DateTime.Now;
					DateTime next = programQueue_.Keys[0] - extraTimeSpan_;
					if (now >= next)
					{
						// ̃L[̃Xg擾AXgׂď
						List<RegisteredProgram> queue = programQueue_.Values[0];
						while (queue.Count > 0)
						{
							// Xg
							RegisteredProgram program = queue[0];
							queue.RemoveAt(0);

							// \ɑΉ錻݂̏ԂA\Jn҂Ȃ
							RecordingStates recordingStates = RecordingStates.Waiting;
							if (programStates_.TryGetValue(program.RegisteredID, out recordingStates) == true)
							{
								if (recordingStates == RecordingStates.Waiting)
								{
									// \^Jn
									StartRecord(program, false);
								}
							}
						}
					}
				}

				// ^C}[ăXPW[
				StartTimer();
			}
		}

		/// <summary>
		/// ^揈sXbh̃Gg|CgłB
		/// </summary>
		/// <param name="argument">^悷\Ǝ蓮Jnǂ̃tOKeyPairValue</param>
		private void RecordingEntry(object argument)
		{
			KeyValuePair<RegisteredProgram, bool> pair = (KeyValuePair<RegisteredProgram, bool>)argument;
			RegisteredProgram program = pair.Key;
			bool isManual = pair.Value;

			Thread.CurrentThread.Name = string.Format("Recording: {0}", program.RegisteredID);

			// ^~Cxgo^
			ManualResetEvent stop = new ManualResetEvent(false);
			lock (this)
			{
				stops_.Add(program.RegisteredID, stop);
			}

			try
			{
				// IZoiI݂͌̎ɌvZKvj
				DateTime realEndTime = program.StartTime + program.Duration + extraTimeSpan_;
				DateTime realStartTime = DateTime.Now;
				TimeSpan realDuration = realEndTime - realStartTime;

				// Xg[WpXtH[}bgĎ擾
				// "{8}_{9}_{10}\{0}_{2:yyyyMMdd}_{2:HHmmss}_{5:yyyyMMdd}_{5:HHmmss}.ts"
				string storagePath = storagePaths_[program.StorageID];
				string filePath = string.Format(storagePath, program.FileName, program.RegisteredID,
					program.StartTime, program.StartTime.ToUniversalTime(), program.Duration,
					realStartTime, realStartTime.ToUniversalTime(), realDuration,
					program.TuningSpace, program.Channel, program.ProgramID,
					program.RegisteredTime, program.RegisteredTime.ToUniversalTime());

				// ݂̃Xg[W̋󂫗eʂsĂꍇA^揈sƂȂI
				long currentSpaces = GetCurrentSpaces(program.StorageID);
				if (currentSpaces < mininumSpaceRequirements_)
				{
					WriteLog(program, "Recorder out of storage space", filePath, realStartTime, realDuration, currentSpaces);
					return;
				}

				// ܂Xg[WpXɃtH_Ȃΐ
				string folderPath = Path.GetDirectoryName(filePath);
				if (Directory.Exists(folderPath) == false)
				{
					Directory.CreateDirectory(folderPath);
				}

				// Xg[WpXɑ΂ăt@CI[v
				using (FileStream fileStream = new FileStream(filePath, FileMode.Create, FileAccess.ReadWrite, FileShare.None))
				{
					// Xg[ڑ
					string url = string.Format("{0}?tuningSpace={1}&channel={2}&programID={3}&locked=true",
						program.TunerStreamUrl, program.TuningSpace, program.Channel, program.ProgramID);
					HttpWebRequest webRequest = (HttpWebRequest)WebRequest.Create(url);
					webRequest.ProtocolVersion = HttpVersion.Version11;
					webRequest.Method = "GET";
					webRequest.Credentials = CredentialCache.DefaultNetworkCredentials;
					webRequest.AllowAutoRedirect = true;
					webRequest.KeepAlive = false;
					webRequest.ReadWriteTimeout = Timeout.Infinite;
					webRequest.Timeout = 15000;
					webRequest.UserAgent = "Recorder; " + this.GetType().Assembly.FullName;

					HttpWebResponse webResponse = (HttpWebResponse)webRequest.GetResponse();
					if (webResponse.StatusCode != HttpStatusCode.OK)
					{
						throw new InvalidDataException(
							"Http response status code is not OK: StatusCode=" + webResponse.StatusCode);
					}

					try
					{
						Stream webStream = webResponse.GetResponseStream();
						byte[] buffer = new byte[bufferSize_];

						if (isManual == true)
						{
							WriteLog(program, "Begin recording with manual start",
								filePath, realStartTime, realDuration, currentSpaces);
						}
						else
						{
							WriteLog(program, "Begin recording",
								filePath, realStartTime, realDuration, currentSpaces);
						}

						// Xg[t@Cɓ]
						bool stopped = false;
						bool finished = false;
						while (true)
						{
							// Iɘ^JnꍇAI܂ł̊Ԋu^CAEgƂĐݒ肷
							int timeout = Timeout.Infinite;
							if (isManual == false)
							{
								DateTime now = DateTime.Now;
								lock (program)
								{
									realEndTime = program.StartTime + program.Duration + extraTimeSpan_;
									realDuration = realEndTime - now;
								}

								// ɏIo߂ĂΏI
								if (now >= realEndTime)
								{
									finished = true;
									break;
								}
								if ((realEndTime - now).TotalMilliseconds > int.MaxValue)
								{
									timeout = int.MaxValue;
								}
							}

							// Xg[̓ǂݎ񓯊Ŏs
							IAsyncResult ar = webStream.BeginRead(buffer, 0, buffer.Length, null, null);
							int read = 0;
							try
							{
								// T[rX̃A{[gA̘^̒~AXg[̎MA^CAEg܂ő҂
								int result = WaitHandle.WaitAny(new WaitHandle[] { abort_, stop, ar.AsyncWaitHandle }, timeout);

								// ^CAEgȂAXg[ǂݎLZĖ߂
								if (result == WaitHandle.WaitTimeout)
								{
									continue;
								}

								// T[rX̃A{[g
								if (result == 0)
								{
									break;
								}

								// ^̒~
								if (result == 1)
								{
									stopped = true;
									break;
								}
							}
							finally
							{
								read = webStream.EndRead(ar);
							}

							// Xg[ǂݎ
							if (read == 0)
							{
								// Xg[T[o[ꂽHiA{[gƂĈj
								break;
							}

							// t@Cɏo͂
							fileStream.Write(buffer, 0, read);

							// ݂̃Xg[W̋󂫗eʂsĂꍇAA{[g
							currentSpaces = GetCurrentSpaces(program.StorageID);
							if (currentSpaces < mininumSpaceRequirements_)
							{
								break;
							}
						}

						long length = fileStream.Length;
						fileStream.Close();

						currentSpaces = GetCurrentSpaces(program.StorageID);
						if (isManual == true)
						{
							WriteLog(program, "Recording stopped", filePath, realStartTime, realDuration, currentSpaces);
						}
						else
						{
							WriteLog(program, "Recording finished", filePath, realStartTime, realDuration, currentSpaces);
						}

						lock (this)
						{
							if (programStates_.ContainsKey(program.RegisteredID) == true)
							{
								if (finished == true)
								{
									programStates_[program.RegisteredID] = RecordingStates.RecordingFnished;
								}
								else if (stopped == true)
								{
									programStates_[program.RegisteredID] = RecordingStates.RecordingStoppedForManual;
								}
								else
								{
									programStates_[program.RegisteredID] = RecordingStates.RecordingAborted;
								}
							}
						}
					}
					finally
					{
						// WebResponse.Close()Stream.Close()̓ubNĂ܂ꍇ
						webRequest.Abort();
					}
				}
			}
			catch (Exception ex)
			{
				lock (this)
				{
					// G[̂߁AA{[gԂƂ
					if (programStates_.ContainsKey(program.RegisteredID) == true)
					{
						programStates_[program.RegisteredID] = RecordingStates.RecordingAborted;
					}
				}

				// OL^
				Logger.Write(ex);
			}
			finally
			{
				// ^~Cxg폜
				lock (this)
				{
					stops_.Remove(program.RegisteredID);
				}
				stop.Close();
			}
		}

		[Serializable]
		private sealed class RegisteredProgramPair
		{
			private readonly RegisteredProgram program_;
			private readonly RecordingStates recordingState_;

			public RegisteredProgramPair(RegisteredProgram program, RecordingStates recordingState)
			{
				program_ = program;
				recordingState_ = recordingState;
			}

			public RegisteredProgram Program
			{
				get
				{
					return program_;
				}
			}

			public RecordingStates RecordingStatus
			{
				get
				{
					return recordingState_;
				}
			}
		}
	}
}
