//////////////////////////////////////////////////////////////////////////
// BonSDK Friio (White/Black) tuner driver library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// BonTunerBase.cpp: Friio`[i[NX
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "Resource.h"
#include "BonSDK.h"
#include "BonTunerBase.h"
#include <TChar.h>


/////////////////////////////////////////////////////////////////////////////
// t@C[J萔ݒ
/////////////////////////////////////////////////////////////////////////////

// obt@ݒ
#define ASYNCBUFFTIME		2UL											// obt@ = 5b
#define ASYNCBUFFNUM		(0x400000UL / TSRECVSIZE * ASYNCBUFFTIME)	// 32MbpsƂ

#define REQRESERVNUM		16UL			// 񓯊NGXg\

// I/ORg[R[h
#define IOCODE_CONTROL		0x00220020UL	// `[iRg[
#define IOCODE_STREAM		0x0022004BUL	// Xg[NGXg


/////////////////////////////////////////////////////////////////////////////
// I/ONGXgL[f[^ێNX
/////////////////////////////////////////////////////////////////////////////

// 񓯊I/OśAobt@𑶑KvB
// ܂A񓯊I/OCxgێKv邽߁ÃNX
// ꌳǗB
class CBonTunerBase::ASYNC_REQ_DATA
{
private:
	CSmartEvent *m_pEvent;
	IMediaData *m_pMediaData;

	ASYNC_REQ_DATA(const ASYNC_REQ_DATA &);
	void operator =(const ASYNC_REQ_DATA &);

public:
	OVERLAPPED OverLapped;
	DWORD dwRxdSize;
	BYTE TxdBuff[64];

	ASYNC_REQ_DATA()
		: m_pEvent(0), dwRxdSize(0), m_pMediaData(new CMediaData(0))
	{
		::ZeroMemory(&OverLapped, sizeof OverLapped);
		::ZeroMemory(TxdBuff, sizeof TxdBuff);

		m_pMediaData->SetSize(TSRECVSIZE);

	}

	~ASYNC_REQ_DATA()
	{
		Close();
		BON_SAFE_RELEASE(m_pMediaData);
	}

	void Initialize(const BYTE *pTxdData, DWORD size)
	{
		Close();

		m_pEvent = new CSmartEvent(true);

		::ZeroMemory(&OverLapped, sizeof OverLapped);
		OverLapped.hEvent = m_pEvent->UnsafeGetHandle();

		::CopyMemory(TxdBuff, pTxdData, size);
		dwRxdSize = 0;
	}

	IMediaData *GetMediaData()
	{
		return m_pMediaData;
	}

	const IKernelObject *GetKernelObject() const
	{
		return m_pEvent;
	}

	void Close()
	{
		if (m_pEvent != 0)
		{
			delete m_pEvent;
			m_pEvent = 0;
		}
	}
};


/////////////////////////////////////////////////////////////////////////////
// Friiop@TS`[i HALC^tF[XNX
/////////////////////////////////////////////////////////////////////////////

HINSTANCE CBonTunerBase::m_hModule = NULL;

const BONGUID CBonTunerBase::GetDeviceType(void) const
{
	// foCX̃^CvԂ
	return ::BON_NAME_TO_GUID(TEXT("IHalTsTuner"));
}

const DWORD CBonTunerBase::GetTotalDeviceNum(void) const
{
	// FriiȏԂ
	return (m_BonEnumerator.EnumDevice(true))? m_BonEnumerator.GetTotalNum() : 0UL;
}

const DWORD CBonTunerBase::GetActiveDeviceNum(void) const
{
	// gpFriio̐Ԃ
	return (m_BonEnumerator.EnumDevice(true))? m_BonEnumerator.GetActiveNum() : 0UL;
}

const DWORD CBonTunerBase::OpenTuner(void)
{
	// UN[Y
	CloseTuner();

	try
	{
		if(!OnOpenTuner()) throw __LINE__;

		// gp̃`[ifoCXpX擾
		const CBonString driverPath = m_BonEnumerator.LockAvailableDevice(m_dwDeviceIndex);
		if(driverPath.IsEmpty()) throw __LINE__;

		// hCoI[v
		if((m_hFriioDriver = ::CreateFile(driverPath,
			GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ | FILE_SHARE_WRITE,
			NULL, OPEN_EXISTING, FILE_FLAG_OVERLAPPED, NULL)) ==
			INVALID_HANDLE_VALUE) throw __LINE__;

		// XbhN
		if(!m_IoReqThread.StartThread(m_pName, this, &CBonTunerBase::IoReqThread)) throw __LINE__;
	}
	catch (...)
	{
		// G[
		CloseTuner();
		return EC_OPEN_FAILED;
	}

	return EC_OPEN_SUCCESS;
}

void CBonTunerBase::CloseTuner(void)
{
	// Xbh~
	m_IoReqThread.EndThread();

	if(m_hFriioDriver != INVALID_HANDLE_VALUE)
	{
		// hCoN[Y
		::CloseHandle(m_hFriioDriver);
		m_hFriioDriver = INVALID_HANDLE_VALUE;
	}

	// hCobNJ
	m_BonEnumerator.ReleaseDevice();

	// `l
	m_dwCurSpace = 0UL;
	m_dwCurChannel = 0UL;

	m_dwDeviceIndex = 0xffffffffUL;
}

const DWORD CBonTunerBase::GetDeviceID(void) const
{
	return m_dwDeviceIndex;
}

const DWORD CBonTunerBase::GetCurSpace(void) const
{
	// ݂̃`[jOԂԂ
	return m_dwCurSpace;
}

const DWORD CBonTunerBase::GetCurChannel(void) const
{
	// ݂̃`lԂ
	return m_dwCurChannel;
}

const bool CBonTunerBase::SetChannel(const DWORD dwSpace, const DWORD dwChannel)
{
	// `[jO͈̓`FbN
	if (EnumChannelName(dwSpace, dwChannel).IsEmpty()) return false;

	if (!OnSetChannel(dwSpace, dwChannel)) return false;

	// `lXV
	m_dwCurSpace = dwSpace;
	m_dwCurChannel = dwChannel;

	return true;
}

const bool CBonTunerBase::SetLnbPower(const bool bEnable)
{
	return false;
}

const bool CBonTunerBase::SetOutputDecoder(IMediaDecoder *pDecoder)
{
	m_pOutputDecoder = pDecoder;
	return true;
}

CBonTunerBase::CBonTunerBase(IBonObject *pOwner, int initSeqID, int endSeqID, LPCTSTR pName)
	: CBonObject(pOwner)
	, m_hFriioDriver(NULL)
	, m_dwCurSpace(0UL)
	, m_dwCurChannel(0UL)
	, m_dwBufferSize(0)
	, m_dwBufferPosition(0)
	, m_pAsyncBuffer(0)
	, m_initSeqID(initSeqID)
	, m_endSeqID(endSeqID)
	, m_pStandardTuningSpace(BON_SAFE_CREATE<IStandardTuningSpace *>(TEXT("CStandardTuningSpace"), 0))
	, m_pOutputDecoder(0)
	, m_dwDeviceIndex(0xffffffffUL)
	, m_IoReqThread()
	, m_pName(pName)
{

}

CBonTunerBase::~CBonTunerBase(void)
{
	CloseTuner();

	BON_SAFE_RELEASE(m_pStandardTuningSpace);
}

const CBonString CBonTunerBase::GetStandardTuningSpaceName(const IStandardTuningSpace::STANDARD_TUNING_SPACES tuningSpace) const
{
	return m_pStandardTuningSpace->GetStandardTuningSpaceName(tuningSpace);
}

const CBonString CBonTunerBase::GetStandardChannelName(const IStandardTuningSpace::STANDARD_TUNING_SPACES tuningSpace, const DWORD dwChannel) const
{
	return m_pStandardTuningSpace->GetStandardChannelName(tuningSpace, dwChannel);
}

const int CBonTunerBase::SendIoRequest(const DWORD dwIoCode, const BYTE *pTxdData, const DWORD dwTxdSize, BYTE *pRxdData, const DWORD dwRxdSize) const
{
	// I[v`FbN
	if (m_hFriioDriver == INVALID_HANDLE_VALUE) return -1;

	// Mpf[^ݒ
	OVERLAPPED OverLapped;
	::ZeroMemory(&OverLapped, sizeof(OverLapped));
	CSmartEvent event(true);
	OverLapped.hEvent = event.UnsafeGetHandle();

	// f[^M
	DWORD dwReturned = 0UL;
	const BOOL bReturn = ::DeviceIoControl(m_hFriioDriver, dwIoCode, const_cast<BYTE *>(pTxdData), dwTxdSize, pRxdData, dwRxdSize, &dwReturned, &OverLapped);

	// G[AMf[^擾
	if (!bReturn)
	{
		if (::GetLastError() != ERROR_IO_PENDING) return -1;
		if (!::GetOverlappedResult(m_hFriioDriver, &OverLapped, &dwReturned, TRUE)) return -1;
	}

	return (int)dwReturned;
}

void CBonTunerBase::SetModuleHandle(const HINSTANCE hModule)
{
	// W[nho^
	m_hModule = hModule;
}

const HINSTANCE CBonTunerBase::GetModuleInstance()
{
	return m_hModule;
}

const bool CBonTunerBase::SendFixedRequest(const UINT nID) const
{
	// \[Xnh擾
	HRSRC hRsrc = ::FindResource(m_hModule, MAKEINTRESOURCE(nID), TEXT("BINARY"));
	if(!hRsrc)return false;

	// \[X[h
	HGLOBAL hGlobal = ::LoadResource(m_hModule, hRsrc);
	if(!hGlobal)return false;

	// |C^擾
	BYTE *pFixedBuff = (BYTE *)::LockResource(hGlobal);
	if(!pFixedBuff)return false;

	// \[XTCY擾
	const DWORD dwIoReqSize = ::SizeofResource(m_hModule, hRsrc);

	// \[Xf[^𑗐M
	DWORD dwXmited = 0UL;
	FIXED_REQ_HEAD *pFixedReqHead;

	while (dwXmited < dwIoReqSize)
	{
		// Rg[擾
		pFixedReqHead = (FIXED_REQ_HEAD *)pFixedBuff;
		pFixedBuff += sizeof(FIXED_REQ_HEAD);

		// Mobt@NA
		BYTE RxdBuff[256];
		::ZeroMemory(RxdBuff, sizeof RxdBuff);

		// f[^M
		if (SendIoRequest(pFixedReqHead->dwIoCode, pFixedBuff, pFixedReqHead->wTxdSize, RxdBuff, pFixedReqHead->wRxdSize) == -1) return false;
		pFixedBuff += pFixedReqHead->wTxdSize;

		// MσTCYXV
		dwXmited += (sizeof(FIXED_REQ_HEAD) + pFixedReqHead->wTxdSize);
	}

	return true;
}

const bool CBonTunerBase::AllocateBuffer(const DWORD dwBufferNum)
{
	ReleaseBuffer();

	m_dwBufferSize = max(dwBufferNum, REQRESERVNUM + 1);
	m_dwBufferPosition = 0;
	m_pAsyncBuffer = new ASYNC_REQ_DATA[m_dwBufferSize];

	return true;
}

void CBonTunerBase::ReleaseBuffer(void)
{
	// NGXgL[J
	while (m_AsyncReqQue.size() > 0)
	{
		m_AsyncReqQue.pop();
	}

	if (m_pAsyncBuffer != 0)
	{
		delete [] m_pAsyncBuffer;
		m_pAsyncBuffer = 0;
		m_dwBufferSize = 0;
		m_dwBufferPosition = 0;
	}
}

void CBonTunerBase::IoReqThread(CSmartThread<CBonTunerBase> &thread, const ISignal &bKillSignal, const ULONG_PTR param)
{
	// V[PXM
	SendFixedRequest(m_initSeqID);

	// ftHg`lݒ
	SetChannel(0UL, 0UL);

	// obt@m
	AllocateBuffer(REQRESERVNUM + 1);

	try
	{
		// hCoɃXg[NGXg𔭍s
		while (true)
		{
			// hCoTSf[^NGXg𔭍s
			PushIoRequest();

			// σf[^΃NGXg
			if (PopIoRequest(bKillSignal) == false)
			{
				break;
			}
		}
	}
	catch (...)
	{
	}

	// IV[PXM
	SendFixedRequest(m_endSeqID);

	// ׂĂ̔񓯊I/OLZ
	// 񓯊I/ÕLŹAfoCXɂsȂB
	// ܂A񓯊I/O𔭍sXbhłpȂB
	// ifriioĂAǂɂĂLZ͏oȂj
	::CancelIo(m_hFriioDriver);

	// obt@J
	ReleaseBuffer();
}

const bool CBonTunerBase::PushIoRequest(void)
{
	// Xg[NGXg
	static const BYTE abyStreamReq[]  = {0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x81, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

	// NGXg҂KɒB܂ŌJԂ
	while (m_AsyncReqQue.size() < REQRESERVNUM)
	{
		// NGXgݒ
		ASYNC_REQ_DATA *pAsyncReq = m_pAsyncBuffer + m_dwBufferPosition;
		pAsyncReq->Initialize(abyStreamReq, sizeof abyStreamReq);
		IMediaData *pMediaData = pAsyncReq->GetMediaData();

		// NGXgs
		if (!::DeviceIoControl(m_hFriioDriver, IOCODE_STREAM,
			pAsyncReq->TxdBuff, sizeof abyStreamReq,
			pMediaData->GetData(), pMediaData->GetSize(),
			NULL, &pAsyncReq->OverLapped))
		{
			if (::GetLastError() != ERROR_IO_PENDING)
			{
				break;
			}
		}

		// L[ɓ
		m_AsyncReqQue.push(pAsyncReq);
		m_dwBufferPosition++;
		if (m_dwBufferPosition >= m_dwBufferSize)
		{
			m_dwBufferPosition = 0;
		}
	}

	return true;
}

const bool CBonTunerBase::PopIoRequest(const ISignal &bKillSignal)
{
	if (m_AsyncReqQue.size() == 0)
	{
		return true;
	}
	ASYNC_REQ_DATA *pAsyncReq = m_AsyncReqQue.front();

	// Cxg҂
	IWaitableObject::WaitResult result = (*bKillSignal.GetKernelObject() || *pAsyncReq->GetKernelObject()).Wait();
	if (result != IWaitableObject::Signal1)
	{
		return false;
	}

	if (m_AsyncReqQue.size() == 0)
	{
		return true;
	}

	m_AsyncReqQue.pop();

	// G[łȂ
	if (::GetOverlappedResult(m_hFriioDriver, &pAsyncReq->OverLapped, &pAsyncReq->dwRxdSize, FALSE))
	{
		// CxgnhN[Y
		pAsyncReq->Close();

		if (pAsyncReq->dwRxdSize > 0)
		{
			if (m_pOutputDecoder != 0)
			{
				IMediaData *pMediaData = pAsyncReq->GetMediaData();
				
				pMediaData->SetSize(pAsyncReq->dwRxdSize);
				m_pOutputDecoder->InputMedia(pMediaData);
			}
		}
	}
	else
	{
		// CxgnhN[Y
		pAsyncReq->Close();
	}

	return true;
}
