//////////////////////////////////////////////////////////////////////////
// BonSDK Media decoder library
// Copyright (c) http://2sen.dip.jp/dtv/
//
// This is part of BonSDK.
// BonSDK comes with ABSOLUTELY NO WARRANTY;
// for details see 'GNU GENERAL PUBLIC LICENSE 2.0'.

// FileReader.cpp: t@C[_fR[_
//
/////////////////////////////////////////////////////////////////////////////

#include "StdAfx.h"
#include "BonSDK.h"
#include "FileReader.h"


/////////////////////////////////////////////////////////////////////////////
// t@C[J萔ݒ
/////////////////////////////////////////////////////////////////////////////

// ftHg̃obt@TCY
#define DEF_BUFFSIZE	0x100000UL							// 1MB
#define DEF_BUFFNUM		8UL									// 250ms  


/////////////////////////////////////////////////////////////////////////////
// t@C[_fR[_
/////////////////////////////////////////////////////////////////////////////

const bool CFileReader::OpenFile(LPCTSTR lpszFileName, const bool bAsyncRead, const DWORD dwBuffSize)
{
	CloseFile();

	try{
		// obt@TCYvZ
		m_dwBuffSize = (dwBuffSize)? dwBuffSize : DEF_BUFFSIZE;
		const DWORD dwReadNum = (bAsyncRead)? DEF_BUFFNUM : 1UL;

		// X}[gt@CCX^X
		if(!(m_pFile = ::BON_SAFE_CREATE<ISmartFile *>(TEXT("CSmartFile"))))throw __LINE__;

		// t@CI[v
		if(!m_pFile->Open(lpszFileName, ISmartFile::MF_READ, m_dwBuffSize))throw __LINE__;

		// t@CTCY`FbN
		if(!m_pFile->GetLength())throw __LINE__;

		// fBAv[TCYTCY
		m_MediaPool.resize(dwReadNum);
		if(m_MediaPool.size() < dwReadNum)throw __LINE__;
		::ZeroMemory(&m_MediaPool[0], sizeof(IMediaData *) * dwReadNum);
			
		// fBAf[^CX^Xm
		for(DWORD dwIndex = 0UL ; dwIndex < dwReadNum ; dwIndex++){
			m_MediaPool[dwIndex] = ::BON_SAFE_CREATE<IMediaData *>(TEXT("CMediaData"));
			if(!m_MediaPool[dwIndex])throw __LINE__;
				
			// obt@m
			m_MediaPool[dwIndex]->SetSize(m_dwBuffSize);
			}
			
		// fBAv[擪Ce[^Zbg
		m_itFreeMedia = m_MediaPool.begin();			
		}
	catch(...){
		// G[
		CloseFile();
		return false;
		}
		
	return true;
}

bool CFileReader::CloseFile(void)
{
	// fR[_Ot~
	StopDecoder();

	// t@CN[Y
	if(m_pFile)m_pFile->Close();

	// X}[gt@CCX^XJ
	BON_SAFE_RELEASE(m_pFile);

	// fBAL[NA
	while(!m_MediaQue.empty())m_MediaQue.pop();

	// fBAf[^CX^XJ
	for(DWORD dwIndex = 0UL ; dwIndex < m_MediaPool.size() ; dwIndex++){
		BON_SAFE_RELEASE(m_MediaPool[dwIndex]);		
		}
	
	// fBAv[J
	m_MediaPool.clear();

	return true;
}

const DWORD CFileReader::ReadSync(const DWORD dwReadSize)
{
	if(!m_pFile || !dwReadSize)return 0UL;
	
	// vTCYm
	if((*m_itFreeMedia)->SetSize(dwReadSize) != dwReadSize)return 0UL;

	// f[^ǂݍ
	const DWORD dwActualSize = m_pFile->ReadData((*m_itFreeMedia)->GetData(), dwReadSize);
	if(!dwActualSize)return 0UL;

	// LTCYݒ
	if((*m_itFreeMedia)->SetSize(dwActualSize) != dwActualSize)return 0UL;

	// fBAf[^o
	return OutputMedia(*m_itFreeMedia)? dwActualSize : 0UL;
}

const bool CFileReader::SetReadPos(const ULONGLONG llPos)
{
	CBlockLock AutoLock(&m_Lock);
	
	// t@CV[N
	return (m_pFile)? m_pFile->SeekPos(llPos) : false;
}

const ULONGLONG CFileReader::GetReadPos(void)
{
	CBlockLock AutoLock(&m_Lock);
	
	// t@C|WVԂ
	return (m_pFile)? m_pFile->GetPos() : 0ULL;
}

const ULONGLONG CFileReader::GetFileLength(void)
{
	CBlockLock AutoLock(&m_Lock);
	
	// t@CTCYԂ
	return (m_pFile)? m_pFile->GetLength() : 0ULL;
}

CFileReader::CFileReader(IBonObject *pOwner)
	: CMediaDecoder(pOwner, 0UL, 1UL)
	, m_pFile(NULL)
	, m_dwBuffSize(DEF_BUFFSIZE)
	, m_ReadThread(new CFlagSignal(1))
	, m_OutputThread(new CFlagSignal(1))
{
	// Ȃ
}

CFileReader::~CFileReader(void)
{
	// t@C
	CloseFile();
}

const bool CFileReader::OnPlay(void)
{
	OnStop();
	
	// fBAL[NA
	while(!m_MediaQue.empty())m_MediaQue.pop();
	
	// fBAv[擪Ce[^Zbg
	m_itFreeMedia = m_MediaPool.begin();

	// XbhJn
	if(m_MediaPool.size()){
		// o̓XbhJn
		if(!m_OutputThread.StartThread(TEXT("CFileReader::OutputThread"), this, &CFileReader::OutputThread, NULL, false))return false;
		
		// ǂݍ݃XbhJn
		if(!m_ReadThread.StartThread(TEXT("CFileReader::ReadThread"), this, &CFileReader::ReadThread)){
			m_OutputThread.EndThread(true);
			}
		}
		
	return true;
}

const bool CFileReader::OnStop(void)
{
	// ǂݍ݃XbhI
	m_ReadThread.EndThread(true);

	// o̓XbhI
	m_OutputThread.EndThread(true);
	
	return true;
}

void CFileReader::ReadThread(CSmartThread<CFileReader> &thread, const ISignal &bKillSignal, const ULONG_PTR param)
{
	DWORD dwIdleNum;

	// JnCxgʒm
	SendDecoderEvent(IFileReader::EID_READ_START);

	bool kill = false;

	while(!kill && !bKillSignal){
		// L[ɋ󂫂t@Cǂݍ
		while(!kill && !bKillSignal){
			{
				// AChL[擾
				CBlockLock AutoLock(&m_Lock);
				dwIdleNum = m_MediaPool.size() - m_MediaQue.size();
			}
			if(!dwIdleNum)break;

			// ǂݍݑOCxgʒm
			SendDecoderEvent(IFileReader::EID_PRE_READ);

			// t@Cǂݍ
			if(!ReadAsync(m_dwBuffSize)){
			
				// t@CI[Ao̓L[̃f[^Sď܂ő҂
				while(m_OutputThread.IsRunning()){
					{
						// L[擾
						CBlockLock AutoLock(&m_Lock);
						dwIdleNum = m_MediaQue.size();
					}
					if(!dwIdleNum)break;
					
					// XbhTXyhAL[ɋ󂫂o܂ő҂
					thread.SuspendThread();
					}

				kill = true;
				break;
				}

			// ǂݍ݌Cxgʒm
			SendDecoderEvent(IFileReader::EID_POST_READ);

			// L[ɒǉ
			CBlockLock AutoLock(&m_Lock);
			m_MediaQue.push(*m_itFreeMedia);
			if(++m_itFreeMedia == m_MediaPool.end())m_itFreeMedia = m_MediaPool.begin();
			
			// o̓XbhW[
			m_OutputThread.ResumeThread();
			}

		// XbhTXyhAL[ɋ󂫂o܂ő҂
		if(!kill && !bKillSignal)thread.SuspendThread();
		}

	// ICxgʒm
	SendDecoderEvent(IFileReader::EID_READ_END);
}

void CFileReader::OutputThread(CSmartThread<CFileReader> &thread, const ISignal &bKillSignal, const ULONG_PTR param)
{
	IMediaData *pMediaData;
	
	while(!bKillSignal){
		while(!bKillSignal){
			{
				// o͑҂L[擾
				CBlockLock AutoLock(&m_Lock);
				pMediaData = (!m_MediaQue.empty())? m_MediaQue.front() : NULL;
			}
			if(!pMediaData)break;
			
			// fBAf[^o
			OutputMedia(pMediaData);

			// L[o
			CBlockLock AutoLock(&m_Lock);
			m_MediaQue.pop();
			
			// ǂݍ݃XbhW[
			m_ReadThread.ResumeThread();
			}
		
		// XbhTXyhAL[Ƀf[^܂ő҂
		if(!bKillSignal)thread.SuspendThread();
		}
}

const DWORD CFileReader::ReadAsync(const DWORD dwReadSize)
{
	if(!m_pFile)return 0UL;
	
	// vTCYm
	if((*m_itFreeMedia)->SetSize(dwReadSize) != dwReadSize)return 0UL;

	// f[^ǂݍ
	const DWORD dwActualSize = m_pFile->ReadData((*m_itFreeMedia)->GetData(), m_dwBuffSize);
	if(!dwActualSize)return 0UL;

	// LTCYݒ
	if((*m_itFreeMedia)->SetSize(dwActualSize) != dwActualSize)return 0UL;

	return dwActualSize;
}
